#include "mqbench.h"

#define FAILED_MQSR(Status) ((MQLONG)(Status)!=MQCC_OK)
#define ASSERT_VALIDHANDLE(handle) assert(handle!=NULL)
#define ASSERT_VALIDPHANDLE(phandle) assert(phandle!=NULL);assert(*phandle!=NULL)

/* internal MQS representation for Queue Handle */
typedef struct
{
  MQHCONN   hConn;
  MQHOBJ    hQueue;
} MQSHANDLEOBJ, *PMQSHANDLEOBJ;

/* global message defines */
MQMD      g_MqMd;
MQPMO     g_MqPmo;
MQGMO     g_MqGmo;
char*     g_pMessageBody = NULL;

void Usage_I(void)
{
  static char* pszThreadStr = "[-t count ] ";
  static char* pszNoThreadStr = "";

	printf ("usage: MQSRBench [-s[e|r] count size] [-r count size] [-q queue] [-c]\n");
  printf ("                 [-x count] %s[-w] [-p]\n",g_MultiThreadSupport ? pszThreadStr : pszNoThreadStr);
  printf ("                 [-z[s|r] senderQ receiverQ] [-?]\n");

	printf ("-s\tsend 'count' express or recoverable messages. size is in bytes\n");
	printf ("-r\treceive 'count' messages. size is in bytes\n");
	printf ("-x\tuse 'count' messages for each sync point\n");
	if( g_MultiThreadSupport ) printf ("-t\tuse threads\n");
	printf ("-q\tqueue name (format QMGR\\DEST-QMGR\\QNAME)\n");
	printf ("-c\treceiver clears the queue prior to running\n");
	printf ("-w\twait for first message before measurement\n");
  printf ("-p\tprints test scenario information prior to running test\n");
  printf ("-z\tsyncs the sender and receiver. if (s)ender is used, the\n");
  printf ("  \treceiver starts after sender has finished. if (r)eceiver is used,\n");
  printf ("  \tthe receiver is running before the sender starts running\n");
  printf ("  \tsync queue names are of the format DEST-QMGR\\QNAME\n");
	printf ("-?\tprint this help\n");
	exit(-1);
}

void InitParams_I(void) {
	g_Params.MsgType = MQPER_NOT_PERSISTENT;
	g_Params.XactType = NO_SYNCPOINT;
	g_Params.I.pQMgrName[0] = '\0';
	g_Params.I.pDestQMgr[0] = '\0';
	g_Params.I.pQueueName[0] = '\0';
	g_Params.I.pSyncSendDestQMgr[0] = '\0';
	g_Params.I.pSyncSendQueueName[0] = '\0';
	g_Params.I.pSyncRecvDestQMgr[0] = '\0';
	g_Params.I.pSyncRecvQueueName[0] = '\0';
}

void ParsePathParams_I(char* pszPathname, QTYPE qType) {
  char *pStr1,*pStr2,*pStr3;

  pStr1 = strtok(pszPathname,"\\");
  if(!pStr1) return;

  pStr2 = strtok(NULL,"\\");
  if(!pStr2) return;

  if(qType == APP_Q) {
    pStr3 = strtok(NULL,"\\");
    if(!pStr3) return;
  }

  switch(qType)
  {
  case APP_Q:
    strcpy(g_Params.I.pQMgrName,pStr1);
    strcpy(g_Params.I.pDestQMgr,pStr2);
    strcpy(g_Params.I.pQueueName,pStr3);
    break;
 
  case SENDSYNC_Q:
    strcpy(g_Params.I.pSyncSendDestQMgr,pStr1);
    strcpy(g_Params.I.pSyncSendQueueName,pStr2);
    break;

  case RECVSYNC_Q:
    strcpy(g_Params.I.pSyncRecvDestQMgr,pStr1);
    strcpy(g_Params.I.pSyncRecvQueueName,pStr2);
    break;
  }
}

int IsValidPathname_I(void) {
	return  g_Params.I.pQMgrName[0] != '\0' &&
	        g_Params.I.pDestQMgr[0] != '\0' &&
	        g_Params.I.pQueueName[0] != '\0' &&
          ( g_Params.SyncMode == NO_SYNC ||
	        (g_Params.I.pSyncSendDestQMgr[0] != '\0' &&
	        g_Params.I.pSyncSendQueueName[0] != '\0' &&
	        g_Params.I.pSyncRecvDestQMgr[0] != '\0' &&
	        g_Params.I.pSyncRecvQueueName[0] != '\0' ) );
}

int ParseTransactionParams_I(int Cur, int ParamNum, char *pszParams[])
{
	if (!IsValidParam (Cur, ParamNum, pszParams))
	{
		printf ("Missing number of messages in sync point.\n");
		Usage_I();
	}

  g_Params.XactType = SYNCPOINT;
	g_Params.ulTransactionNum = atol(pszParams[Cur+1]);
	if (g_Params.ulTransactionNum == 0)
	{
		printf("Invalid number of transaction.\n");
		Usage_I();
	}

	return Cur + 2;
}

char* XACTTYPE2STR_I(DWORD dwXactType) {
  char* psz;
  switch (dwXactType) {
    DEF2STR(NO_SYNCPOINT);
    DEF2STR(SYNCPOINT);
    default: psz = "UNKNOWN_XACTTYPE";
  }
  return psz;
}

void InitTest_I(void)
{
	SetMessageProps_I();
}

void SetMessageProps_I(void)
{
  MQMD      MqMd = {MQMD_DEFAULT};
  MQPMO     MqPmo = {MQPMO_DEFAULT};
  MQGMO     MqGmo = {MQGMO_DEFAULT};

	/*
	// Initialize global structures to default values
	*/
  memcpy(&g_MqMd,&MqMd,sizeof(MQMD));
  memcpy(&g_MqPmo,&MqPmo,sizeof(MQPMO));
  memcpy(&g_MqGmo,&MqGmo,sizeof(MQGMO));
    
	/*
	// Set the message body buffer
	*/
	g_pMessageBody = (char*) malloc(g_Params.ulMsgSize);
	
	if (g_Params.Operation == SEND)
	{
		/*
		// Build message body
		*/
		memset(g_pMessageBody, 'a', g_Params.ulMsgSize);
	}

	/*
	// Set message persistence
	*/
  g_MqMd.Persistence = 
      g_Params.MsgType == EXPRESS ? MQPER_NOT_PERSISTENT :
                                    MQPER_PERSISTENT;
  /*
  // Set The Get and Put Options 
  */
  g_MqGmo.WaitInterval = MQWI_UNLIMITED;

  if( g_Params.XactType == SYNCPOINT ) {
    g_MqPmo.Options |= MQPMO_SYNCPOINT;
    g_MqGmo.Options |= MQGMO_SYNCPOINT | MQGMO_WAIT;
  }
  else {
    g_MqPmo.Options |= MQPMO_NO_SYNCPOINT;
    g_MqGmo.Options |= MQGMO_NO_SYNCPOINT | MQGMO_WAIT;
  }
}

MQHOBJ OpenQueue(MQHCONN hConn, char* pQMgrName, char* pQName, MQLONG lOptions) {
  MQOD      MqOd = {MQOD_DEFAULT};
  MQHOBJ    hQueue;
  MQLONG    lCompCode,lReason;

  strcpy(MqOd.ObjectName,pQName);
  strcpy(MqOd.ObjectQMgrName,pQMgrName);

  MQOPEN(hConn,&MqOd,lOptions,&hQueue,&lCompCode,&lReason); 
	if (FAILED_MQSR(lCompCode)) 
  {
		ErrorExit("MQOPEN failed", lReason);
  }

  return hQueue;
}
 
HANDLE GetQueueHandle_I(void)
{
  MQLONG        lCompCode,lReason;
	MQLONG        lOptions;
  PMQSHANDLEOBJ phObj;

  phObj = (PMQSHANDLEOBJ) malloc(sizeof(MQSHANDLEOBJ));
  ASSERT_VALIDHANDLE(phObj);

  /* connect to queue manager */
  MQCONN(g_Params.I.pQMgrName,&(phObj->hConn),&lCompCode,&lReason);
	if (FAILED_MQSR(lCompCode)) 
	{
		ErrorExit("MQCONN failed", lReason);
	}

	if (g_Params.Operation==SEND)
	{
		lOptions = MQOO_OUTPUT;
	}
	else
	{
		lOptions = MQOO_INPUT_SHARED;
	}

  phObj->hQueue = OpenQueue(phObj->hConn,g_Params.I.pDestQMgr,g_Params.I.pQueueName,lOptions);

  return (HANDLE)phObj;
}

PTRANSOBJ GetTransactionPointer_I(void)
{
  return NULL;
}

void SendMsg_I(PTRANSOBJ pXact,HANDLE hQueue) {
  MQMD          MqMd;
  MQLONG        lCompCode,lReason;
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;

  ASSERT_VALIDHANDLE(phObj);

  memcpy(&MqMd,&g_MqMd,sizeof(MQMD));
  MQPUT(phObj->hConn,phObj->hQueue,&MqMd,&g_MqPmo,g_Params.ulMsgSize,g_pMessageBody,
        &lCompCode,&lReason);

	if (FAILED_MQSR(lCompCode)) 
	{
    CloseQueue_I(&hQueue);
		ErrorExit("MQPUT failed", lReason);
	}
}

void RecvMsg_I(PTRANSOBJ pXact, HANDLE hQueue) {
  MQMD          MqMd;
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;
  MQLONG        lDataLength;
  MQLONG        lCompCode,lReason;

  ASSERT_VALIDHANDLE(phObj);

  memcpy(&MqMd,&g_MqMd,sizeof(MQMD));
  MQGET(phObj->hConn,phObj->hQueue,&MqMd,&g_MqGmo,g_Params.ulMsgSize,g_pMessageBody,&lDataLength,
        &lCompCode,&lReason);

	if (FAILED_MQSR(lCompCode)) 
	{
    CloseQueue_I(&hQueue);
		ErrorExit("MQGET failed", lReason);
	}
}

void CommitTrans_I(PTRANSOBJ pXact, HANDLE hQueue) {
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;
  MQLONG        lCompCode,lReason;

  ASSERT_VALIDHANDLE(phObj);

  MQCMIT(phObj->hConn,&lCompCode,&lReason);
	if (FAILED_MQSR(lCompCode)) 
  {
    CloseQueue_I(&hQueue);
	  ErrorExit("MQCMIT failed", lReason);
	}
}

void CloseQueue_I(HANDLE* phQueue) {
  PMQSHANDLEOBJ phObj;
  MQLONG        lCompCode,lReason;

  ASSERT_VALIDHANDLE(phQueue);
  phObj = (PMQSHANDLEOBJ)*phQueue;
  ASSERT_VALIDHANDLE(phObj);

  MQCLOSE(phObj->hConn,&phObj->hQueue,MQCO_NONE,&lCompCode,&lReason);
  MQDISC(&phObj->hConn,&lCompCode,&lReason);
  free(phObj);

  *phQueue = NULL;
}

void CleanupTest_I(void)
{
  free(g_pMessageBody);
  g_pMessageBody = NULL;
}

void PrintQueuename_I(QTYPE qType) {
  switch(qType) {
    case APP_Q:
    	printf("Queue Name:             %s\\%s connected to %s\n", 
              g_Params.I.pDestQMgr,g_Params.I.pQueueName,
              g_Params.I.pQMgrName);
      break;
   
    case SENDSYNC_Q:
    	printf("Send Sync Queue Name:   %s\\%s connected to %s\n", 
              g_Params.I.pSyncSendDestQMgr,g_Params.I.pSyncSendQueueName,
              g_Params.I.pQMgrName);
      break;

    case RECVSYNC_Q:
      printf("Recv Sync Queue Name:   %s\\%s connected to %s\n", 
              g_Params.I.pSyncRecvDestQMgr,g_Params.I.pSyncRecvQueueName,
              g_Params.I.pQMgrName);
      break;
  }
}

DWORD ClearQueue_I(HANDLE hQueue) {
	MQHOBJ        hReceiveQ;
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;
  MQLONG        lDataLength;
  MQLONG        lCompCode = MQCC_OK,lReason;
  ULONG         ulBodySize = MAX_PATH;
  char          pMsgBody[MAX_PATH];
  DWORD         dwCount = 0;

  ASSERT_VALIDHANDLE(phObj);

	hReceiveQ = OpenQueue(phObj->hConn,g_Params.I.pDestQMgr,g_Params.I.pQueueName,
                        MQOO_INPUT_SHARED);

  while(lCompCode == MQCC_OK) {
    MQMD          MqMd = {MQMD_DEFAULT};
    MQGMO         MqGmo = {MQGMO_DEFAULT};

    /* initialize Get Message Options */
    MqGmo.Options |= MQGMO_NO_SYNCPOINT | MQGMO_NO_WAIT | MQGMO_ACCEPT_TRUNCATED_MSG;

    MQGET(phObj->hConn,hReceiveQ,&MqMd,&MqGmo,ulBodySize,pMsgBody,&lDataLength,
          &lCompCode,&lReason);
    if (lCompCode == MQCC_OK || lReason == MQRC_TRUNCATED_MSG_ACCEPTED) {
      lCompCode = MQCC_OK;
      dwCount++;
    }
  }
  
  MQCLOSE(phObj->hConn,&hReceiveQ,MQCO_NONE,&lCompCode,&lReason);

  return dwCount;
}

void OpenSyncQueues_I(HANDLE hQueue, SYNCDATA* pSyncData) {
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;
  MQLONG        lSyncSendOptions;
  MQLONG        lSyncRecvOptions;

  ASSERT_VALIDHANDLE(phObj);

	if (g_Params.Operation==SEND)
	{
		lSyncSendOptions = MQOO_INPUT_SHARED;
		lSyncRecvOptions = MQOO_OUTPUT;
	}
	else
	{
		lSyncSendOptions = MQOO_OUTPUT;
		lSyncRecvOptions = MQOO_INPUT_SHARED;
	}
			
  pSyncData->hSendSyncQ = (HANDLE) OpenQueue( phObj->hConn,g_Params.I.pSyncSendDestQMgr,
                                              g_Params.I.pSyncSendQueueName,lSyncSendOptions);
  pSyncData->hRecvSyncQ = (HANDLE) OpenQueue( phObj->hConn,g_Params.I.pSyncRecvDestQMgr,
                                              g_Params.I.pSyncRecvQueueName,lSyncRecvOptions);
}

void CloseSyncQueues_I(HANDLE hQueue, SYNCDATA* pSyncData) {
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;
  MQLONG        lCompCode,lReason;

  ASSERT_VALIDHANDLE(phObj);

  MQCLOSE(phObj->hConn,(PMQHOBJ)&(pSyncData->hSendSyncQ),MQCO_NONE,&lCompCode,&lReason);
  MQCLOSE(phObj->hConn,(PMQHOBJ)&(pSyncData->hRecvSyncQ),MQCO_NONE,&lCompCode,&lReason);
}

void GetSyncMessage_I(HANDLE hQueue, SYNCDATA* pSyncData,char* pMsgId, char* pCorrId, 
                      SYNCOPERATION SyncOper) {
  MQHOBJ        hReceiveQ;
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;
  MQLONG        lDataLength;
  MQLONG        lCompCode,lReason;
  ULONG         ulBodySize = SYNCOPERTEXTSIZE;
  char          pMsgBody[SYNCOPERTEXTSIZE];
  BOOL          fCorrectMsg = FALSE;

  ASSERT_VALIDHANDLE(phObj);

  /* determine which queue handle we will be receiving from */
	if (g_Params.Operation==SEND)
	{
		hReceiveQ = (MQHOBJ)pSyncData->hSendSyncQ;
	}
	else
	{
		hReceiveQ = (MQHOBJ)pSyncData->hRecvSyncQ;
	}

  while(!fCorrectMsg) {
    MQMD          MqMd = {MQMD_DEFAULT};
    MQGMO         MqGmo = {MQGMO_DEFAULT};

    /* initialize Get Message Options */
    MqGmo.WaitInterval = SYNC_TIMEOUT;
    MqGmo.Options |= MQGMO_NO_SYNCPOINT | MQGMO_WAIT;

    MQGET(phObj->hConn,hReceiveQ,&MqMd,&MqGmo,ulBodySize,pMsgBody,&lDataLength,
          &lCompCode,&lReason);

	  if (FAILED_MQSR(lCompCode)) 
	  {
      CloseSyncQueues_I(hQueue, pSyncData);
      CloseQueue_I(&hQueue);
		  ErrorExit("MQGET failed", lReason);
	  }

    /* check to see if this is the correct message */
    if( !memcmp(SYNCOPERTEXT[SyncOper],pMsgBody,ulBodySize) && 
        (!pCorrId || (pCorrId && (!(char*)memcmp(pCorrId,MqMd.CorrelId,sizeof(MQBYTE24)))))) {
      /* copy the MsgId if it was needed */
      if(pMsgId) {
        /* check if the message originated from the FalconMQ Bridge */
        /* if so, ignore the first 4 bytes */
        if(!strncmp(pMsgId,"FQ2Q",4)) {
          memset(pMsgId,' ',sizeof(MQBYTE24));
          memcpy(pMsgId,MqMd.MsgId + 4,sizeof(MQBYTE24) - 4);
        } else {
          memcpy(pMsgId,MqMd.MsgId,sizeof(MQBYTE24));
        }
      }
      fCorrectMsg = TRUE;
    }
  } /* end while */
}

void PutSyncMessage_I(HANDLE hQueue, SYNCDATA* pSyncData,char* pMsgId, char* pCorrId, 
                      SYNCOPERATION SyncOper) {
  MQHOBJ        hSendQ;
  MQMD          MqMd = {MQMD_DEFAULT};
  MQPMO         MqPmo = {MQPMO_DEFAULT};
  MQLONG        lCompCode,lReason;
  PMQSHANDLEOBJ phObj = (PMQSHANDLEOBJ)hQueue;

  ASSERT_VALIDHANDLE(phObj);

  /* determine which queue handle we will be sending to */
	if (g_Params.Operation==SEND)
	{
		hSendQ = (MQHOBJ)pSyncData->hRecvSyncQ;
	}
	else
	{
		hSendQ = (MQHOBJ)pSyncData->hSendSyncQ;
	}

  /* initialize Put Message Options */
  MqPmo.Options |= MQPMO_NO_SYNCPOINT;

  /* copy the CorrId if it is being used */
  if( pCorrId ) {
    memcpy(MqMd.CorrelId,pCorrId,sizeof(MQBYTE24));
  }

  /* set the expiry time for the message.  Note expiry is 100's of a second */
  MqMd.Expiry = SYNC_MSG_LIFETIME / 10;

  MQPUT(phObj->hConn,hSendQ,&MqMd,&MqPmo,SYNCOPERTEXTSIZE,SYNCOPERTEXT[SyncOper],
        &lCompCode,&lReason);

	if (FAILED_MQSR(lCompCode)) 
	{
    CloseSyncQueues_I(hQueue, pSyncData);
    CloseQueue_I(&hQueue);
		ErrorExit("MQPUT failed", lReason);
	}

  /* copy the MsgId if it was needed */
  if(pMsgId) {
    memcpy(pMsgId,MqMd.MsgId,sizeof(MQBYTE24));
  }
}
